//-----------------------------------------------------------------------------
// Copyright (c) 2007-2008 dhpoware. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
//
// This Windows OpenGL application demonstrates offscreen rendering to a
// texture and then applying that texture to a screen aligned full screen quad.
//
// In this demo we use offscreen rendering to render our scene to a texture
// with the same dimensions as the screen. Once the scene is rendered to
// this offscreen texture the intention is to then apply some kind of post
// processing filter to it. Some examples of post processing filters include
// full scene motion blur, tone mapping (for HDR rendering), and specular
// bloom. After the post processing is applied the offscreen texture is then
// applied to a full screen quad and displayed on the screen. We don't actually 
// perform any kind of post processing in this demo. This demo simply sets up
// the required infrastructure for offscreen rendering and post processing.
//
// This demo uses the GL_EXT_framebuffer_object (FBO) extension to handle
// rendering to a texture. This is preferred to the older PBuffer extension
// since FBOs share the same OpenGL context as the one used by the window of
// your application. PBuffers require their own context.
//
// Until recently FBOs did not support multisampling. Two new extensions -
// GL_EXT_framebuffer_multisample and GL_EXT_framebuffer_blit - extend the
// existing FBO framework by providing support for multisample rendering. The 
// GL_EXT_framebuffer_multisample extension allows the allocation of render
// buffer images that can be used as a multisample buffer. However no mechanism
// is provided for creating multisample texture images. What this means is that
// to have FBOs that support multisampling we need to use two FBOs. The first
// FBO will have 2 multisample render buffer images: one for the color buffer
// and the other for the depth buffer. This is the FBO that the scene will be
// rendered to. The rendered scene will be multisampled. The second FBO
// contains no multisample render buffer images. It just has the offscreen
// texture directly attached to one of the FBO's color attachment points. To
// transfer the pixels in the multisample color render buffer image to the
// offscreen texture we use the GL_EXT_framebuffer_blit extension. The
// GL_EXT_framebuffer_blit extensions modifies FBOs by splitting their binding
// points into separate DRAW and READ bindings. This allows the direct copying
// of the contents of one FBO to another. Once the copy operation is complete
// the offscreen texture will contain a copy of the pixel data from the
// first FBO's multisample color render buffer image. 
//
// Screen resolutions are not powers of two. As a result this demo uses the
// OpenGL texture rectangle extension for non power of two texture support. We
// opted to use GL_ARB_texture_rectangle rather than the newer
// GL_ARB_texture_non_power_of_two extension since the former is more likely to
// be supported on older hardware. The latter is now part of the OpenGL 2.0
// specification. One of the limitations of the GL_ARB_texture_rectangle
// extension is that is does not support mipmaps. This is not a problem for this
// demo since we are only ever going to use the offscreen texture on a screen
// aligned full screen quad. This demo contains code paths for both extensions.
// By default it uses the texture rectangle path since it is more likely to be
// supported on older hardware. The GL_ARB_texture_non_power_of_two code path
// can be enabled by uncommenting a defined constant.
//
// This demo uses the following OpenGL extensions:
//  GL_EXT_framebuffer_object
//  GL_EXT_framebuffer_multisample
//  GL_EXT_framebuffer_blit
//  GL_ARB_texture_non_power_of_two
//  GL_ARB_texture_rectangle
//  WGL_ARB_extensions_string
//  WGL_EXT_swap_control
//
// The multisample extensions to FBOs are quite new and the best reference on
// how to use these new extensions is the NVIDIA OpenGL SDK 10 Simple
// Framebuffer Object Sample.
// http://developer.download.nvidia.com/SDK/10/Samples/simple_framebuffer_object.zip)
//
//-----------------------------------------------------------------------------

#if !defined(WIN32_LEAN_AND_MEAN)
#define WIN32_LEAN_AND_MEAN
#endif

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <cassert>
#include <cmath>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <string>

#if defined(_DEBUG)
#include <crtdbg.h>
#endif

#include "GL_EXT_framebuffer_object.h"
#include "gl_font.h"

//-----------------------------------------------------------------------------
// Constants.
//-----------------------------------------------------------------------------

#define APP_TITLE "OpenGL Offscreen Rendering Demo"

// Windows Vista compositing support.
#if !defined(PFD_SUPPORT_COMPOSITION)
#define PFD_SUPPORT_COMPOSITION 0x00008000
#endif

// Uncomment the following line to use GL_ARB_texture_non_power_of_two instead
// of GL_ARB_texture_rectangle.
//#define USE_GL_ARB_TEXTURE_NON_POWER_OF_TWO

// GL_ARB_texture_rectangle
#define GL_TEXTURE_RECTANGLE_ARB          0x84F5
#define GL_TEXTURE_BINDING_RECTANGLE_ARB  0x84F6
#define GL_PROXY_TEXTURE_RECTANGLE_ARB    0x84F7
#define GL_MAX_RECTANGLE_TEXTURE_SIZE_ARB 0x84F8

#define CAMERA_FOVY  60.0f
#define CAMERA_ZFAR  10.0f
#define CAMERA_ZNEAR 0.1f

#define MOUSE_ORBIT_SPEED 0.30f     // 0 = SLOWEST, 1 = FASTEST
#define MOUSE_DOLLY_SPEED 0.02f     // same as above...but much more sensitive
#define MOUSE_TRACK_SPEED 0.005f    // same as above...but much more sensitive

#define CUBE_SIZE       1.0f
#define CUBE_SIZE_HALF (CUBE_SIZE * 0.5f)

//-----------------------------------------------------------------------------
// Types.
//-----------------------------------------------------------------------------

struct Vertex
{
    float pos[3];
    float normal[3];
};

//-----------------------------------------------------------------------------
// Globals.
//-----------------------------------------------------------------------------

HWND                g_hWnd;
HDC                 g_hDC;
HGLRC               g_hRC;
HINSTANCE           g_hInstance;
int                 g_framesPerSecond;
int                 g_windowWidth;
int                 g_windowHeight;
float               g_heading;
float               g_pitch;
float               g_cameraPos[3] = {0.0f, 0.0f, CUBE_SIZE * 2.0f};
float               g_targetPos[3];
bool                g_isFullScreen;
bool                g_hasFocus;
bool                g_enableVerticalSync;
bool                g_displayHelp;
int                 g_fboWidth;
int                 g_fboHeight;
int                 g_fboMaxRenderBufferSize;
int                 g_fboMaxSamples;
int                 g_fboSamples;
GLuint              g_fbo;
GLuint              g_fboResolveTarget;
GLuint              g_fboColorBuffer;
GLuint              g_fboDepthBuffer;
GLuint              g_offscreenTexture;
GLuint              g_fullScreenQuadDisplayList;
GLFont              g_font;

Vertex g_cube[24] =
{
    // +Z face
    {-CUBE_SIZE_HALF, -CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   0.0f,  0.0f,  1.0f},
    { CUBE_SIZE_HALF, -CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   0.0f,  0.0f,  1.0f},
    { CUBE_SIZE_HALF,  CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   0.0f,  0.0f,  1.0f},
    {-CUBE_SIZE_HALF,  CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   0.0f,  0.0f,  1.0f},

    // -Z face
    { CUBE_SIZE_HALF, -CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   0.0f,  0.0f, -1.0f},
    {-CUBE_SIZE_HALF, -CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   0.0f,  0.0f, -1.0f},
    {-CUBE_SIZE_HALF,  CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   0.0f,  0.0f, -1.0f},
    { CUBE_SIZE_HALF,  CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   0.0f,  0.0f, -1.0f},

    // +Y face
    {-CUBE_SIZE_HALF,  CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   0.0f,  1.0f,  0.0f},
    { CUBE_SIZE_HALF,  CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   0.0f,  1.0f,  0.0f},
    { CUBE_SIZE_HALF,  CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   0.0f,  1.0f,  0.0f},
    {-CUBE_SIZE_HALF,  CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   0.0f,  1.0f,  0.0f},

    // -Y face
    {-CUBE_SIZE_HALF, -CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   0.0f, -1.0f,  0.0f},
    { CUBE_SIZE_HALF, -CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   0.0f, -1.0f,  0.0f},
    { CUBE_SIZE_HALF, -CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   0.0f, -1.0f,  0.0f},
    {-CUBE_SIZE_HALF, -CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   0.0f, -1.0f,  0.0f},

    // +X face
    { CUBE_SIZE_HALF, -CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   1.0f,  0.0f,  0.0f},
    { CUBE_SIZE_HALF, -CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   1.0f,  0.0f,  0.0f},
    { CUBE_SIZE_HALF,  CUBE_SIZE_HALF, -CUBE_SIZE_HALF,   1.0f,  0.0f,  0.0f},
    { CUBE_SIZE_HALF,  CUBE_SIZE_HALF,  CUBE_SIZE_HALF,   1.0f,  0.0f,  0.0f},

    // -X face
    {-CUBE_SIZE_HALF, -CUBE_SIZE_HALF, -CUBE_SIZE_HALF,  -1.0f,  0.0f,  0.0f},
    {-CUBE_SIZE_HALF, -CUBE_SIZE_HALF,  CUBE_SIZE_HALF,  -1.0f,  0.0f,  0.0f},
    {-CUBE_SIZE_HALF,  CUBE_SIZE_HALF,  CUBE_SIZE_HALF,  -1.0f,  0.0f,  0.0f},
    {-CUBE_SIZE_HALF,  CUBE_SIZE_HALF, -CUBE_SIZE_HALF,  -1.0f,  0.0f,  0.0f}
};

//-----------------------------------------------------------------------------
// Functions Prototypes.
//-----------------------------------------------------------------------------

void    CheckFramebufferStatus();
void    Cleanup();
void    CleanupApp();
HWND    CreateAppWindow(const WNDCLASSEX &wcl, const char *pszTitle);
void    DrawFrame();
void    DrawFullScreenQuad();
void    DrawScene();
void    DrawText();
void    EnableVerticalSync(bool enableVerticalSync);
bool    ExtensionSupported(const char *pszExtensionName);
float   GetElapsedTimeInSeconds();
bool    Init();
void    InitApp();
void    InitFBO();
void    InitFBOWithMultiSampling(int samples);
void    InitGL();
void    Log(const char *pszMessage);
void    ProcessMouseInput(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
void    SetProcessorAffinity();
void    ToggleFullScreen();
void    UpdateFrame(float elapsedTimeSec);
void    UpdateFrameRate(float elapsedTimeSec);
LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

//-----------------------------------------------------------------------------
// Functions.
//-----------------------------------------------------------------------------

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
#if defined _DEBUG
    _CrtSetDbgFlag(_CRTDBG_LEAK_CHECK_DF | _CRTDBG_ALLOC_MEM_DF);
    _CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_FILE);
    _CrtSetReportFile(_CRT_ASSERT, _CRTDBG_FILE_STDERR);
#endif

    MSG msg = {0};
    WNDCLASSEX wcl = {0};

    wcl.cbSize = sizeof(wcl);
    wcl.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
    wcl.lpfnWndProc = WindowProc;
    wcl.cbClsExtra = 0;
    wcl.cbWndExtra = 0;
    wcl.hInstance = g_hInstance = hInstance;
    wcl.hIcon = LoadIcon(0, IDI_APPLICATION);
    wcl.hCursor = LoadCursor(0, IDC_ARROW);
    wcl.hbrBackground = 0;
    wcl.lpszMenuName = 0;
    wcl.lpszClassName = "GLWindowClass";
    wcl.hIconSm = 0;

    if (!RegisterClassEx(&wcl))
        return 0;

    g_hWnd = CreateAppWindow(wcl, APP_TITLE);

    if (g_hWnd)
    {
        SetProcessorAffinity();

        if (Init())
        {
            ShowWindow(g_hWnd, nShowCmd);
            UpdateWindow(g_hWnd);

            while (true)
            {
                while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
                {
                    if (msg.message == WM_QUIT)
                        break;

                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }

                if (msg.message == WM_QUIT)
                    break;

                if (g_hasFocus)
                {
                    UpdateFrame(GetElapsedTimeInSeconds());
                    DrawFrame();
                    SwapBuffers(g_hDC);
                }
                else
                {
                    WaitMessage();
                }
            }
        }

        Cleanup();
        UnregisterClass(wcl.lpszClassName, hInstance);
    }

    return static_cast<int>(msg.wParam);
}

LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg)
    {
    case WM_ACTIVATE:
        switch (wParam)
        {
        default:
            break;

        case WA_ACTIVE:
        case WA_CLICKACTIVE:
            g_hasFocus = true;
            break;

        case WA_INACTIVE:
            if (g_isFullScreen)
                ShowWindow(hWnd, SW_MINIMIZE);
            g_hasFocus = false;
            break;
        }
        break;

    case WM_CHAR:
        switch (static_cast<int>(wParam))
        {
        case VK_ESCAPE:
            PostMessage(hWnd, WM_CLOSE, 0, 0);
            break;

        case 'h':
        case 'H':
            g_displayHelp = !g_displayHelp;
            break;

        case 'v':
        case 'V':
            EnableVerticalSync(!g_enableVerticalSync);
            break;

        default:
            break;
        }
        break;

    case WM_DESTROY:
        PostQuitMessage(0);
        return 0;

    case WM_SIZE:
        g_windowWidth = static_cast<int>(LOWORD(lParam));
        g_windowHeight = static_cast<int>(HIWORD(lParam));
        break;

    case WM_SYSKEYDOWN:
        if (wParam == VK_RETURN)
            ToggleFullScreen();
        break;

    default:
        ProcessMouseInput(hWnd, msg, wParam, lParam);
        break;
    }

    return DefWindowProc(hWnd, msg, wParam, lParam);
}

void CheckFramebufferStatus()
{
    switch (glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT))
    {
    case GL_FRAMEBUFFER_COMPLETE_EXT:
        break;

    case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
        throw std::runtime_error("Setup FBO failed. Unsupported framebuffer format.");

    case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
        throw std::runtime_error("Setup FBO failed. Missing attachment.");

    case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
        throw std::runtime_error("Setup FBO failed. Duplicate attachment.");

    case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
        throw std::runtime_error("Setup FBO failed. Attached images must have the same dimensions.");

    case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
        throw std::runtime_error("Setup FBO failed. Attached images must have the same format.");

    case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
        throw std::runtime_error("Setup FBO failed. Missing draw buffer.");

    case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
        throw std::runtime_error("Setup FBO failed. Missing read buffer.");

    case GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_EXT:
        throw std::runtime_error("Setup FBO failed. Attached images must have the same number of samples.");

    default:
        throw std::runtime_error("Setup FBO failed. Fatal error.");
    }
}

void Cleanup()
{
    CleanupApp();

    if (g_hDC)
    {
        if (g_hRC)
        {
            wglMakeCurrent(g_hDC, 0);
            wglDeleteContext(g_hRC);
            g_hRC = 0;
        }

        ReleaseDC(g_hWnd, g_hDC);
        g_hDC = 0;
    }
}

void CleanupApp()
{
    if (g_fbo)
    {
        glDeleteFramebuffersEXT(1, &g_fbo);
        g_fbo = 0;
    }
    
    if (g_fboResolveTarget)
    {
        glDeleteFramebuffersEXT(1, &g_fboResolveTarget);
        g_fboResolveTarget = 0;
    }

    if (g_fboColorBuffer)
    {
        glDeleteRenderbuffersEXT(1, &g_fboColorBuffer);
        g_fboColorBuffer = 0;
    }

    if (g_fboDepthBuffer)
    {
        glDeleteRenderbuffersEXT(1, &g_fboDepthBuffer);
        g_fboDepthBuffer = 0;
    }

    if (g_offscreenTexture)
    {
        glDeleteTextures(1, &g_offscreenTexture);
        g_offscreenTexture = 0;
    }

    if (g_fullScreenQuadDisplayList)
    {
        glDeleteLists(g_fullScreenQuadDisplayList, 1);
        g_fullScreenQuadDisplayList = 0;
    }

    g_font.destroy();
}

HWND CreateAppWindow(const WNDCLASSEX &wcl, const char *pszTitle)
{
    // Create a window that is centered on the desktop. It's exactly 1/4 the
    // size of the desktop. Don't allow it to be resized.

    DWORD wndExStyle = WS_EX_OVERLAPPEDWINDOW;
    DWORD wndStyle = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
                     WS_MINIMIZEBOX | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;

    HWND hWnd = CreateWindowEx(wndExStyle, wcl.lpszClassName, pszTitle,
                    wndStyle, 0, 0, 0, 0, 0, 0, wcl.hInstance, 0);

    if (hWnd)
    {
        int screenWidth = GetSystemMetrics(SM_CXSCREEN);
        int screenHeight = GetSystemMetrics(SM_CYSCREEN);
        int halfScreenWidth = screenWidth / 2;
        int halfScreenHeight = screenHeight / 2;
        int left = (screenWidth - halfScreenWidth) / 2;
        int top = (screenHeight - halfScreenHeight) / 2;
        RECT rc = {0};

        SetRect(&rc, left, top, left + halfScreenWidth, top + halfScreenHeight);
        AdjustWindowRectEx(&rc, wndStyle, FALSE, wndExStyle);
        MoveWindow(hWnd, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, TRUE);

        GetClientRect(hWnd, &rc);
        g_windowWidth = rc.right - rc.left;
        g_windowHeight = rc.bottom - rc.top;
    }

    return hWnd;
}

void DrawFrame()
{
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);

    // Render to the offscreen texture.
    // When using multisample rendering we're actually rendering to the
    // multisample FBO and then copying the pixels to the FBO containing the
    // offscreen texture. When not using multisample rendering we render
    // directly to the FBO containing the offscreen texture.
    
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, g_fbo);
    glViewport(0, 0, g_fboWidth, g_fboHeight);   
    
    glClearColor(0.3f, 0.5f, 0.9f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    DrawScene();
    DrawText();
    
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

    if (g_fboSamples > 0)
    {
        // Multisample rendering so copy the pixel data in the multisample
        // color render buffer image to the FBO containing the offscreen
        // texture image.

        glBindFramebufferEXT(GL_READ_FRAMEBUFFER_EXT, g_fbo);
        glBindFramebufferEXT(GL_DRAW_FRAMEBUFFER_EXT, g_fboResolveTarget);
        glBlitFramebufferEXT(0, 0, g_fboWidth, g_fboHeight,
                             0, 0, g_fboWidth, g_fboHeight,
                             GL_COLOR_BUFFER_BIT, GL_NEAREST);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    }

    // At this point we now have our scene fully rendered to our offscreen
    // texture 'g_offscreenTexture'. This is where you would perform any
    // post processing to the offscreen texture.

    // Finally to display the offscreen texture to the screen we draw a screen
    // aligned full screen quad and attach the offscreen texture to it.

    glViewport(0, 0, g_windowWidth, g_windowHeight);   
    
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    DrawFullScreenQuad();
}

void DrawFullScreenQuad()
{
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    glPushAttrib(GL_CURRENT_BIT);

    glDisable(GL_LIGHTING);
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

#if defined(USE_GL_ARB_TEXTURE_NON_POWER_OF_TWO)
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, g_offscreenTexture);

    if (g_fullScreenQuadDisplayList)
    {
        glCallList(g_fullScreenQuadDisplayList);
    }
    else
    {
        g_fullScreenQuadDisplayList = glGenLists(1);
        glNewList(g_fullScreenQuadDisplayList, GL_COMPILE_AND_EXECUTE);

        glBegin(GL_QUADS);
            glTexCoord2f(0.0f, 0.0f);
            glVertex3f(-1.0f, -1.0f, -1.0f);

            glTexCoord2f(1.0f, 0.0f);
            glVertex3f(1.0f, -1.0f, -1.0f);

            glTexCoord2f(1.0f, 1.0f);
            glVertex3f(1.0f, 1.0f, -1.0f);

            glTexCoord2f(0, 1.0f);
            glVertex3f(-1.0f, 1.0f, -1.0f);
        glEnd();

        glEndList();
    }   

    glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);
#else
    glEnable(GL_TEXTURE_RECTANGLE_ARB);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, g_offscreenTexture);

    if (g_fullScreenQuadDisplayList)
    {
        glCallList(g_fullScreenQuadDisplayList);
    }
    else
    {
        g_fullScreenQuadDisplayList = glGenLists(1);
        glNewList(g_fullScreenQuadDisplayList, GL_COMPILE_AND_EXECUTE);

        glBegin(GL_QUADS);
            glTexCoord2i(0, 0);
            glVertex3f(-1.0f, -1.0f, -1.0f);

            glTexCoord2i(g_fboWidth, 0);
            glVertex3f(1.0f, -1.0f, -1.0f);
            
            glTexCoord2i(g_fboWidth, g_fboHeight);
            glVertex3f(1.0f, 1.0f, -1.0f);
            
            glTexCoord2i(0, g_fboHeight);
            glVertex3f(-1.0f, 1.0f, -1.0f);
        glEnd();

        glEndList();
    }   

    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
    glDisable(GL_TEXTURE_RECTANGLE_ARB);
#endif
    
    glPopAttrib();
    glPopMatrix();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    glMatrixMode(GL_MODELVIEW);
}

void DrawScene()
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(CAMERA_FOVY,
        static_cast<float>(g_fboWidth) / static_cast<float>(g_fboHeight),
        CAMERA_ZNEAR, CAMERA_ZFAR);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(g_cameraPos[0], g_cameraPos[1], g_cameraPos[2],
        g_targetPos[0], g_targetPos[1], g_targetPos[2],
        0.0f, 1.0f, 0.0f);

    float matAmbient[4] = {0.19225f, 0.19225f, 0.19225f, 1.0f};
    float matDiffuse[4] = {0.50754f, 0.50754f, 0.50754f, 1.0f};
    float matSpecular[4] = {0.508273f, 0.508273f, 0.508273f, 1.0f};
    float matShininess = 51.2f;
  
    glMaterialfv(GL_FRONT, GL_AMBIENT, matAmbient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, matDiffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, matSpecular);
    glMaterialf(GL_FRONT, GL_SHININESS, matShininess);

    glRotatef(g_pitch, 1.0f, 0.0f, 0.0f);
    glRotatef(g_heading, 0.0f, 1.0f, 0.0f);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY);

    glNormalPointer(GL_FLOAT, sizeof(Vertex), g_cube[0].normal);
    glVertexPointer(3, GL_FLOAT, sizeof(Vertex), g_cube[0].pos);

    glDrawArrays(GL_QUADS, 0, sizeof(g_cube) / sizeof(g_cube[0]));

    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);   

    glDisable(GL_LIGHT0);
    glDisable(GL_LIGHTING);
}

void DrawText()
{
    static RECT rcClient;
    static char szBuffer[512];

    if (g_displayHelp)
    {
        sprintf(szBuffer,
            "Left mouse click and drag to translate the cube\n"
            "Middle mouse click and drag to zoom in and out of the cube\n"
            "Right mouse click and drag to rotate the cube\n\n"
            "Press V to enable/disable vertical sync\n"
            "Press ALT + ENTER to toggle full screen\n"
            "Press ESC to exit\n\n"
            "Press H to hide help");
    }
    else
    {
        sprintf(szBuffer,
            "FPS: %d\n"
            "Vertical sync: %s\n"
            "Window size: %d x %d\n"
            "FBO size: %d x %d\n"
            "FBO samples: %d%s\n\n"
            "Press H to display help",
            g_framesPerSecond,
            ((g_enableVerticalSync) ? "enabled" : "disabled"),
            g_windowWidth, g_windowHeight,
            g_fboWidth, g_fboHeight,
            g_fboSamples, ((g_fboSamples > 0) ? "x MSAA" : ""));
    }
    
    g_font.begin();
    g_font.setColor(1.0f, 1.0f, 0.0f);
    g_font.drawText(1, 1, szBuffer);
    g_font.end();
}

void EnableVerticalSync(bool enableVerticalSync)
{
    // WGL_EXT_swap_control.

    typedef BOOL (WINAPI * PFNWGLSWAPINTERVALEXTPROC)(GLint);

    static PFNWGLSWAPINTERVALEXTPROC wglSwapIntervalEXT =
        reinterpret_cast<PFNWGLSWAPINTERVALEXTPROC>(
        wglGetProcAddress("wglSwapIntervalEXT"));

    if (wglSwapIntervalEXT)
    {
        wglSwapIntervalEXT(enableVerticalSync ? 1 : 0);
        g_enableVerticalSync = enableVerticalSync;
    }
}

bool ExtensionSupported(const char *pszExtensionName)
{
    static const char *pszGLExtensions = 0;
    static const char *pszWGLExtensions = 0;

    if (!pszGLExtensions)
        pszGLExtensions = reinterpret_cast<const char *>(glGetString(GL_EXTENSIONS));

    if (!pszWGLExtensions)
    {
        // WGL_ARB_extensions_string.

        typedef const char *(WINAPI * PFNWGLGETEXTENSIONSSTRINGARBPROC)(HDC);

        PFNWGLGETEXTENSIONSSTRINGARBPROC wglGetExtensionsStringARB =
            reinterpret_cast<PFNWGLGETEXTENSIONSSTRINGARBPROC>(
            wglGetProcAddress("wglGetExtensionsStringARB"));

        if (wglGetExtensionsStringARB)
            pszWGLExtensions = wglGetExtensionsStringARB(g_hDC);
    }

    if (!strstr(pszGLExtensions, pszExtensionName))
    {
        if (!strstr(pszWGLExtensions, pszExtensionName))
            return false;
    }

    return true;
}

float GetElapsedTimeInSeconds()
{
    // Returns the elapsed time (in seconds) since the last time this function
    // was called. This elaborate setup is to guard against large spikes in
    // the time returned by QueryPerformanceCounter().

    static const int MAX_SAMPLE_COUNT = 50;

    static float frameTimes[MAX_SAMPLE_COUNT];
    static float timeScale = 0.0f;
    static float actualElapsedTimeSec = 0.0f;
    static INT64 freq = 0;
    static INT64 lastTime = 0;
    static int sampleCount = 0;
    static bool initialized = false;

    INT64 time = 0;
    float elapsedTimeSec = 0.0f;

    if (!initialized)
    {
        initialized = true;
        QueryPerformanceFrequency(reinterpret_cast<LARGE_INTEGER*>(&freq));
        QueryPerformanceCounter(reinterpret_cast<LARGE_INTEGER*>(&lastTime));
        timeScale = 1.0f / freq;
    }

    QueryPerformanceCounter(reinterpret_cast<LARGE_INTEGER*>(&time));
    elapsedTimeSec = (time - lastTime) * timeScale;
    lastTime = time;

    if (fabsf(elapsedTimeSec - actualElapsedTimeSec) < 1.0f)
    {
        memmove(&frameTimes[1], frameTimes, sizeof(frameTimes) - sizeof(frameTimes[0]));
        frameTimes[0] = elapsedTimeSec;

        if (sampleCount < MAX_SAMPLE_COUNT)
            ++sampleCount;
    }

    actualElapsedTimeSec = 0.0f;

    for (int i = 0; i < sampleCount; ++i)
        actualElapsedTimeSec += frameTimes[i];

    if (sampleCount > 0)
        actualElapsedTimeSec /= sampleCount;

    return actualElapsedTimeSec;
}

bool Init()
{
    try
    {
        InitGL();
        InitApp();
        return true;
    }
    catch (const std::exception &e)
    {
        std::ostringstream msg;

        msg << "Application initialization failed!" << std::endl << std::endl;
        msg << e.what();

        Log(msg.str().c_str());
        return false;
    }
}

void InitApp()
{
    // Check for FBO support.

    if (!ExtensionSupported("GL_EXT_framebuffer_object"))
        throw std::runtime_error("Required extension not supported: GL_EXT_framebuffer_object.");
    
    glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE_EXT, &g_fboMaxRenderBufferSize);
    
    // Check for FBO multisample rendering support.

    g_fboMaxSamples = 0;

    if (ExtensionSupported("GL_EXT_framebuffer_multisample"))
    {
        if (ExtensionSupported("GL_EXT_framebuffer_blit"))
            glGetIntegerv(GL_MAX_SAMPLES_EXT, &g_fboMaxSamples);
        else
            Log("GL_EXT_framebuffer_blit is not supported.\nMultisampling disabled.");
    }
    else
    {
        Log("GL_EXT_framebuffer_multisample is not supported.\nMultisampling disabled.");
    }

    // Check for non power of two texture support.

#if defined(USE_GL_ARB_TEXTURE_NON_POWER_OF_TWO)
    if (!ExtensionSupported("GL_ARB_texture_non_power_of_two"))
        throw std::runtime_error("Required extension not supported: GL_ARB_texture_non_power_of_two.");
#else
    if (!ExtensionSupported("GL_ARB_texture_rectangle"))
        throw std::runtime_error("Required extension not supported: GL_ARB_texture_rectangle.");
#endif

    // Setup framebuffer object.

    if (g_fboMaxSamples > 0)
        InitFBOWithMultiSampling(g_fboMaxSamples);
    else
        InitFBO();

    // Setup fonts.

    if (!g_font.create("Arial", 10, GLFont::BOLD))
        throw std::runtime_error("Failed to create font.");
}

void InitFBO()
{
    // Make the framebuffer object (FBO) the same size as the current desktop.
    // We're relying on the non-power-of-two (NPOT) texture extension here.

    g_fboWidth = GetSystemMetrics(SM_CXSCREEN);
    g_fboHeight = GetSystemMetrics(SM_CYSCREEN);

    // Create the FBO and bind it. We're not supporting multisample rendering
    // here. So that means two things: we only need a single FBO and everything
    // attaches directly to that single FBO.

    glGenFramebuffersEXT(1, &g_fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, g_fbo);

    // Create a texture for the FBO to render to. Since this is for offscreen
    // rendering we don't need mipmaps and we clamp the texture at its edges.
    
#if defined(USE_GL_ARB_TEXTURE_NON_POWER_OF_TWO)
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, &g_offscreenTexture);
    glBindTexture(GL_TEXTURE_2D, g_offscreenTexture);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, g_fboWidth, g_fboHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, g_offscreenTexture, 0);
    glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);
#else
    glEnable(GL_TEXTURE_RECTANGLE_ARB);
    glGenTextures(1, &g_offscreenTexture);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, g_offscreenTexture);
    glTexParameterf(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA8, g_fboWidth, g_fboHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_ARB, g_offscreenTexture, 0);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
    glDisable(GL_TEXTURE_RECTANGLE_ARB);
#endif
    
    // Create a depth render buffer image for the FBO.

    glGenRenderbuffersEXT(1, &g_fboDepthBuffer);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, g_fboDepthBuffer);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, g_fboWidth, g_fboHeight);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, g_fboDepthBuffer);

    // Verify that the FBO is 'complete'.
    // Once the FBO is complete it is ready for us to render to.

    CheckFramebufferStatus();
    
    // Verify that the FBO dimensions are of the expected size.

    int actualWidth = 0;
    int actualHeight = 0;

    glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_WIDTH_EXT, &actualWidth);
    glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_HEIGHT_EXT, &actualHeight);

    assert(actualWidth == g_fboWidth);
    assert(actualHeight == g_fboHeight);

    // Unbind the FBO and switch back to normal window rendering.

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

void InitFBOWithMultiSampling(int samples)
{
    // Make the framebuffer object (FBO) the same size as the current desktop.
    // We're relying on the non-power-of-two (NPOT) texture extension here.

    g_fboWidth = GetSystemMetrics(SM_CXSCREEN);
    g_fboHeight = GetSystemMetrics(SM_CYSCREEN);

    // Create the first of our two FBOs.
    //
    // The first FBO will be the resolve target. This will store the final
    // offscreen texture that will be applied to the full screen quad. Remember
    // that the only reason why we use this resolve target FBO is because FBOs
    // don't support the creation of multisample texture images. We're using
    // the resolve target FBO as a means to get a copy of the pixel data stored
    // in the proper multisample color render buffer image attached to the
    // second FBO (that supports multisample rendering).
    //
    // The resolve target FBO is just a plain FBO. Just attach the offscreen
    // texture to a color attachment point. We don't need a depth buffer or a
    // stencil buffer.
    
    glGenFramebuffersEXT(1, &g_fboResolveTarget);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, g_fboResolveTarget);
                
#if defined(USE_GL_ARB_TEXTURE_NON_POWER_OF_TWO)
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, &g_offscreenTexture);
    glBindTexture(GL_TEXTURE_2D, g_offscreenTexture);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, g_fboWidth, g_fboHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, g_offscreenTexture, 0);
    glBindTexture(GL_TEXTURE_2D, 0);
    glDisable(GL_TEXTURE_2D);
#else
    glEnable(GL_TEXTURE_RECTANGLE_ARB);
    glGenTextures(1, &g_offscreenTexture);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, g_offscreenTexture);
    glTexParameterf(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA8, g_fboWidth, g_fboHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_ARB, g_offscreenTexture, 0);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
    glDisable(GL_TEXTURE_RECTANGLE_ARB);
#endif
    
    // Verify that the first FBO is 'complete'. Once the FBO is complete it is
    // ready for us to use.

    CheckFramebufferStatus();

    // Create the second of our two FBOs.
    //
    // The second FBO will support multisample rendering. This is the actual
    // FBO that our application will render to. As mentioned above the
    // multisample rendering extensions to the FBO framework doesn't allow
    // multisample texture images from being created. That's why for the second
    // FBO we need to create two render buffer images: a color buffer and a
    // depth buffer. Both are multisampled.
    //
    // When creating the multisample render buffer images the samples argument
    // of the glRenderbufferStorageMultisampleEXT() function is a request for
    // the desired minimum number of samples. The actual number of samples
    // allocated for the render buffer image is implementation dependent. But
    // it is guaranteed to be greater than or equal to the requested sample
    // count and no more than the next larger sample count supported by the
    // implementation. We include an assertion in the code below just to
    // emphasize this fact.
    
    int actualWidth = 0;
    int actualHeight = 0;

    glGenFramebuffersEXT(1, &g_fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, g_fbo);

    // Create the multisample color render buffer image and attach it to the
    // second FBO.

    glGenRenderbuffersEXT(1, &g_fboColorBuffer);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, g_fboColorBuffer);
    glRenderbufferStorageMultisampleEXT(GL_RENDERBUFFER_EXT, samples, GL_RGBA8, g_fboWidth, g_fboHeight);
    glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_SAMPLES_EXT, &g_fboSamples);
    assert(g_fboSamples >= samples);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_RENDERBUFFER_EXT, g_fboColorBuffer);

    // Create the multisample depth render buffer image and attach it to the
    // second FBO.

    glGenRenderbuffersEXT(1, &g_fboDepthBuffer);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, g_fboDepthBuffer);
    glRenderbufferStorageMultisampleEXT(GL_RENDERBUFFER_EXT, samples, GL_DEPTH_COMPONENT, g_fboWidth, g_fboHeight);
    glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_SAMPLES_EXT, &g_fboSamples);
    assert(g_fboSamples >= samples);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, g_fboDepthBuffer);

    // Verify that the second FBO is 'complete'. Once the FBO is complete it is
    // ready for us to render to.

    CheckFramebufferStatus();

    // Verify that the second FBO dimensions are of the expected size.

    glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_WIDTH_EXT, &actualWidth);
    glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_HEIGHT_EXT, &actualHeight);

    assert(actualWidth == g_fboWidth);
    assert(actualHeight == g_fboHeight);

    // Unbind the FBO and switch back to normal window rendering.

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

void InitGL()
{
    if (!(g_hDC = GetDC(g_hWnd)))
        throw std::runtime_error("GetDC() failed.");

    PIXELFORMATDESCRIPTOR pfd = {0};
    OSVERSIONINFO osvi = {0};

    pfd.nSize = sizeof(pfd);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 24;
    pfd.cDepthBits = 16;
    pfd.iLayerType = PFD_MAIN_PLANE;

    osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

    if (!GetVersionEx(&osvi))
        throw std::runtime_error("GetVersionEx() failed.");

    // When running under Windows Vista or later support desktop composition.
    if (osvi.dwMajorVersion > 6 || (osvi.dwMajorVersion == 6 && osvi.dwMinorVersion >= 0))
        pfd.dwFlags |=  PFD_SUPPORT_COMPOSITION;

    int pf = ChoosePixelFormat(g_hDC, &pfd);

    if (!SetPixelFormat(g_hDC, pf, &pfd))
        throw std::runtime_error("SetPixelFormat() failed.");

    if (!(g_hRC = wglCreateContext(g_hDC)))
        throw std::runtime_error("wglCreateContext() failed.");

    if (!wglMakeCurrent(g_hDC, g_hRC))
        throw std::runtime_error("wglMakeCurrent() failed.");

    EnableVerticalSync(false);
}

void Log(const char *pszMessage)
{
    MessageBox(0, pszMessage, "Error", MB_ICONSTOP);
}

void ProcessMouseInput(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    // Use the left mouse button to track the camera.
    // Use the middle mouse button to dolly the camera.
    // Use the right mouse button to orbit the camera.

    enum CameraMode {CAMERA_NONE, CAMERA_TRACK, CAMERA_DOLLY, CAMERA_ORBIT};

    static CameraMode cameraMode = CAMERA_NONE;
    static POINT ptMousePrev = {0};
    static POINT ptMouseCurrent = {0};
    static int mouseButtonsDown = 0;
    static float dx = 0.0f;
    static float dy = 0.0f;

    switch (msg)
    {
    case WM_LBUTTONDOWN:
        cameraMode = CAMERA_TRACK;
        ++mouseButtonsDown;
        SetCapture(hWnd);
        ptMousePrev.x = static_cast<int>(static_cast<short>(LOWORD(lParam)));
        ptMousePrev.y = static_cast<int>(static_cast<short>(HIWORD(lParam)));
        ClientToScreen(hWnd, &ptMousePrev);
        break;

    case WM_RBUTTONDOWN:
        cameraMode = CAMERA_ORBIT;
        ++mouseButtonsDown;
        SetCapture(hWnd);
        ptMousePrev.x = static_cast<int>(static_cast<short>(LOWORD(lParam)));
        ptMousePrev.y = static_cast<int>(static_cast<short>(HIWORD(lParam)));
        ClientToScreen(hWnd, &ptMousePrev);
        break;

    case WM_MBUTTONDOWN:
        cameraMode = CAMERA_DOLLY;
        ++mouseButtonsDown;
        SetCapture(hWnd);
        ptMousePrev.x = static_cast<int>(static_cast<short>(LOWORD(lParam)));
        ptMousePrev.y = static_cast<int>(static_cast<short>(HIWORD(lParam)));
        ClientToScreen(hWnd, &ptMousePrev);
        break;

    case WM_MOUSEMOVE:
        ptMouseCurrent.x = static_cast<int>(static_cast<short>(LOWORD(lParam)));
        ptMouseCurrent.y = static_cast<int>(static_cast<short>(HIWORD(lParam)));
        ClientToScreen(hWnd, &ptMouseCurrent);

        switch (cameraMode)
        {
        case CAMERA_TRACK:
            dx = static_cast<float>(ptMouseCurrent.x - ptMousePrev.x);
            dx *= MOUSE_TRACK_SPEED;

            dy = static_cast<float>(ptMouseCurrent.y - ptMousePrev.y);
            dy *= MOUSE_TRACK_SPEED;

            g_cameraPos[0] -= dx;
            g_cameraPos[1] += dy;

            g_targetPos[0] -= dx;
            g_targetPos[1] += dy;

            break;

        case CAMERA_DOLLY:
            dy = static_cast<float>(ptMouseCurrent.y - ptMousePrev.y);
            dy *= MOUSE_DOLLY_SPEED;

            g_cameraPos[2] -= dy;

            if (g_cameraPos[2] < (CUBE_SIZE + CAMERA_ZNEAR))
                g_cameraPos[2] = CUBE_SIZE + CAMERA_ZNEAR;

            if (g_cameraPos[2] > (CAMERA_ZFAR - CUBE_SIZE))
                g_cameraPos[2] = CAMERA_ZFAR - CUBE_SIZE;

            break;

        case CAMERA_ORBIT:
            dx = static_cast<float>(ptMouseCurrent.x - ptMousePrev.x);
            dx *= MOUSE_ORBIT_SPEED;

            dy = static_cast<float>(ptMouseCurrent.y - ptMousePrev.y);
            dy *= MOUSE_ORBIT_SPEED;

            g_heading += dx;
            g_pitch += dy;

            if (g_pitch > 90.0f)
                g_pitch = 90.0f;

            if (g_pitch < -90.0f)
                g_pitch = -90.0f;

            break;
        }

        ptMousePrev.x = ptMouseCurrent.x;
        ptMousePrev.y = ptMouseCurrent.y;
        break;

    case WM_LBUTTONUP:
    case WM_RBUTTONUP:
    case WM_MBUTTONUP:
        if (--mouseButtonsDown <= 0)
        {
            mouseButtonsDown = 0;
            cameraMode = CAMERA_NONE;
            ReleaseCapture();
        }
        else
        {
            if (wParam & MK_LBUTTON)
                cameraMode = CAMERA_TRACK;
            else if (wParam & MK_RBUTTON)
                cameraMode = CAMERA_ORBIT;
            else if (wParam & MK_MBUTTON)
                cameraMode = CAMERA_DOLLY;
        }
        break;

    default:
        break;
    }
}

void SetProcessorAffinity()
{
    // Assign the current thread to one processor. This ensures that timing
    // code runs on only one processor, and will not suffer any ill effects
    // from power management.
    //
    // Based on DXUTSetProcessorAffinity() function from the DXUT framework.

    DWORD_PTR dwProcessAffinityMask = 0;
    DWORD_PTR dwSystemAffinityMask = 0;
    HANDLE hCurrentProcess = GetCurrentProcess();

    if (!GetProcessAffinityMask(hCurrentProcess, &dwProcessAffinityMask, &dwSystemAffinityMask))
        return;

    if (dwProcessAffinityMask)
    {
        // Find the lowest processor that our process is allowed to run against.

        DWORD_PTR dwAffinityMask = (dwProcessAffinityMask & ((~dwProcessAffinityMask) + 1));

        // Set this as the processor that our thread must always run against.
        // This must be a subset of the process affinity mask.

        HANDLE hCurrentThread = GetCurrentThread();

        if (hCurrentThread != INVALID_HANDLE_VALUE)
        {
            SetThreadAffinityMask(hCurrentThread, dwAffinityMask);
            CloseHandle(hCurrentThread);
        }
    }

    CloseHandle(hCurrentProcess);
}

void ToggleFullScreen()
{
    static DWORD savedExStyle;
    static DWORD savedStyle;
    static RECT rcSaved;

    g_isFullScreen = !g_isFullScreen;

    if (g_isFullScreen)
    {
        // Moving to full screen mode.

        savedExStyle = GetWindowLong(g_hWnd, GWL_EXSTYLE);
        savedStyle = GetWindowLong(g_hWnd, GWL_STYLE);
        GetWindowRect(g_hWnd, &rcSaved);

        SetWindowLong(g_hWnd, GWL_EXSTYLE, 0);
        SetWindowLong(g_hWnd, GWL_STYLE, WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS);
        SetWindowPos(g_hWnd, HWND_TOPMOST, 0, 0, 0, 0,
            SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED | SWP_SHOWWINDOW);

        g_windowWidth = GetSystemMetrics(SM_CXSCREEN);
        g_windowHeight = GetSystemMetrics(SM_CYSCREEN);

        SetWindowPos(g_hWnd, HWND_TOPMOST, 0, 0,
            g_windowWidth, g_windowHeight, SWP_SHOWWINDOW);
    }
    else
    {
        // Moving back to windowed mode.

        SetWindowLong(g_hWnd, GWL_EXSTYLE, savedExStyle);
        SetWindowLong(g_hWnd, GWL_STYLE, savedStyle);
        SetWindowPos(g_hWnd, HWND_NOTOPMOST, 0, 0, 0, 0,
            SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED | SWP_SHOWWINDOW);

        g_windowWidth = rcSaved.right - rcSaved.left;
        g_windowHeight = rcSaved.bottom - rcSaved.top;

        SetWindowPos(g_hWnd, HWND_NOTOPMOST, rcSaved.left, rcSaved.top,
            g_windowWidth, g_windowHeight, SWP_SHOWWINDOW);
    }
}

void UpdateFrame(float elapsedTimeSec)
{
    UpdateFrameRate(elapsedTimeSec);
}

void UpdateFrameRate(float elapsedTimeSec)
{
    static float accumTimeSec = 0.0f;
    static int frames = 0;

    accumTimeSec += elapsedTimeSec;

    if (accumTimeSec > 1.0f)
    {
        g_framesPerSecond = frames;

        frames = 0;
        accumTimeSec = 0.0f;
    }
    else
    {
        ++frames;
    }
}